create table users (
  id serial,
  name varchar(200),
  password varchar(200), 

  primary key (id),
  unique (name)
);

create table mailboxes (
  id bigserial,
  user_id integer not null,
  name varchar(200) not null, 

  uid_validity integer not null default 1,
  next_uid integer not null default 1,
  next_modseq bigint not null default 1,
  message_count integer not null default 0,
  seen_count integer not null default 0,
  deleted_count integer not null default 0,
  last_nonrecent_uid integer not null default 0,

  primary key (id),
  foreign key (user_id) references users(id),
  unique (user_id, name)
);

create table messages (
  id bigserial,
  mailbox_id bigint not null,
  uid integer not null,

  seen boolean not null default false,
  answered boolean not null default false,
  flagged boolean not null default false,
  deleted boolean not null default false,
  draft boolean not null default false,
  modseq bigint not null default 0,
  
  primary key (id),
  foreign key (mailbox_id) references mailboxes(id),
  unique (mailbox_id, uid)
);
create index message_modseq on messages(mailbox_id, modseq);

create table message_bodies (
  message_id bigint not null,
  header bytea not null,
  body bytea not null,

  primary key (message_id),
  foreign key (message_id) references messages(id) on delete cascade
);

create or replace function message_insert_trigger() returns trigger as '
begin
  -- this also locks insertions to this mailbox
  update mailboxes set message_count = message_count + 1
    where id = new.mailbox_id;

  if new.seen then
    update mailboxes set seen_count = seen_count + 1
      where id = new.mailbox_id;
  end if;
  if new.deleted then
    update mailboxes set deleted_count = deleted_count + 1
      where id = new.mailbox_id;
  end if;

  if new.uid is null then
    select next_uid into new.uid from mailboxes
      where id = new.mailbox_id;
    update mailboxes set next_uid = next_uid + 1
      where id = new.mailbox_id;
  end if;
  return new;
end;
' language 'plpgsql';
create trigger message_insert_trigger
  before insert on messages for each row
  execute procedure message_insert_trigger();

create or replace function message_update_trigger() returns trigger as '
declare
  v_count integer;
begin
  if old.modseq = new.modseq and
     (old.seen <> new.seen or old.answered <> new.answered or
      old.flagged <> new.flagged or old.deleted <> new.deleted or
      old.draft <> new.draft) then
    -- changing message flags, update modseq
    update mailboxes set next_modseq = next_modseq + 1
      where id = new.mailbox_id;
    select next_modseq - 1 into new.modseq from mailboxes
      where id = new.mailbox_id;

    if new.seen <> old.seen then
      if new.seen then
        v_count := 1;
      else
        v_count := -1;
      end if;
      update mailboxes set seen_count = seen_count + v_count
	where id = new.mailbox_id;
    end if;
    if new.deleted <> old.deleted then
      if new.deleted then
        v_count := 1;
      else
        v_count := -1;
      end if;
      update mailboxes set deleted_count = deleted_count + v_count
	where id = new.mailbox_id;
    end if;
  end if;
  return new;
end;
' language 'plpgsql';
create trigger message_update_trigger
  before update on messages for each row
  execute procedure message_update_trigger();

create or replace function message_delete_trigger() returns trigger as '
begin
  update mailboxes set message_count = message_count - 1
    where id = old.mailbox_id;

  if old.seen then
    update mailboxes set seen_count = seen_count - 1
      where id = old.mailbox_id;
  end if;
  if old.deleted then
    update mailboxes set deleted_count = deleted_count - 1
      where id = old.mailbox_id;
  end if;
  return old;
end;
' language 'plpgsql';
create trigger message_delete_trigger
  before delete on messages for each row
  execute procedure message_delete_trigger();

insert into users (name, password) values ('tss', 'password');
insert into mailboxes (user_id, name) values (1, 'INBOX');
