/* Copyright (c) 2004-2006 Timo Sirainen */

#include "lib.h"
#include "str.h"
#include "var-expand.h"
#include "istream-internal.h"
#include "sql-api.h"
#include "sql-istream.h"

#define SQL_BLOCK_SIZE (1024*128)

struct sql_istream {
	struct istream_private istream;

	struct sql_db *db;

	char *query_template;
	bool querying;
};

static void _close(struct iostream_private *stream ATTR_UNUSED)
{
}

static void _destroy(struct iostream_private *stream)
{
	struct sql_istream *sstream = (struct sql_istream *)stream;

	i_free(sstream->query_template);
	i_free(sstream->istream.w_buffer);
}

static int
handle_result(struct istream_private *stream, struct sql_result *result)
{
	const unsigned char *value;
	size_t size;
	int ret;

	if ((ret = sql_result_next_row(result)) <= 0) {
		stream->istream.stream_errno = EIO;
		stream->istream.eof = TRUE;
		return -1;
	}
	value = sql_result_get_field_value_binary(result, 0, &size);

	i_assert(stream->pos + size <= stream->buffer_size);
	memcpy(stream->w_buffer + stream->pos, value, size);
	stream->pos += size;

	sql_result_free(result);
	return 0;
}

static void sql_istream_query_callback(struct sql_result *result,
				       struct sql_istream *sstream)
{
	(void)handle_result(&sstream->istream, result);
	sstream->querying = FALSE;
}

static ssize_t _read(struct istream_private *stream)
{
	struct sql_istream *sstream = (struct sql_istream *)stream;
	const struct var_expand_table *variables;
	struct sql_result *result;
	string_t *query;
	uoff_t offset;
	size_t size;
	ssize_t ret;

	if (stream->istream.closed)
		return -1;

	offset = stream->istream.v_offset + (stream->pos - stream->skip);
	if (offset >= (uoff_t)stream->statbuf.st_size) {
		stream->istream.eof = TRUE;
		return -1;
	}
	if (sstream->querying)
		return 0;

	stream->istream.stream_errno = 0;

	if (stream->pos == stream->buffer_size) {
		if (stream->skip > 0) {
			/* remove the unused bytes from beginning of buffer */
                        i_stream_compress(stream);
		} else if (stream->max_buffer_size == 0 ||
			   stream->buffer_size < stream->max_buffer_size) {
			/* buffer is full - grow it */
			i_stream_grow_buffer(stream, SQL_BLOCK_SIZE);
		}

		if (stream->pos == stream->buffer_size)
			return -2; /* buffer full */
	}

	size = I_MIN(stream->buffer_size - stream->pos,
		     stream->statbuf.st_size - offset);

	t_push();
	variables = var_expand_table_build('o', dec2str(offset),
					   's', dec2str(size),
					   '\0');

	query = t_str_new(256);
	var_expand(query, sstream->query_template, variables);

	if (stream->istream.blocking) {
		result = sql_query_s(sstream->db, str_c(query));
		if (handle_result(stream, result) < 0)
			ret = -1;
		else
			ret = size;
	} else {
		sstream->querying = TRUE;
		sql_query(sstream->db, str_c(query),
			  sql_istream_query_callback, sstream);
		ret = 0;
	}
	t_pop();

	return ret;
}

static void _seek(struct istream_private *stream, uoff_t v_offset,
		  bool mark ATTR_UNUSED)
{
	stream->istream.stream_errno = 0;
	stream->istream.v_offset = v_offset;
	stream->skip = stream->pos = 0;
}

struct istream *
sql_istream_create(struct sql_db *db, bool async, unsigned long long data_size,
		   const char *query_template, size_t max_buffer_size)
{
	struct sql_istream *sstream;

	sstream = i_new(struct sql_istream, 1);
	sstream->db = db;
	sstream->query_template = i_strdup(query_template);

	sstream->istream.iostream.close = _close;
	sstream->istream.iostream.destroy = _destroy;

	sstream->istream.istream.blocking = !async;
	sstream->istream.max_buffer_size = max_buffer_size;
	sstream->istream.read = _read;
	sstream->istream.seek = _seek;

	(void)i_stream_create(&sstream->istream, -1, 0);

	sstream->istream.statbuf.st_size = data_size;
	return &sstream->istream.istream;
}
