/* Copyright (C) 2007 Timo Sirainen */

#include "lib.h"
#include "str.h"
#include "sql-api.h"
#include "imap-match.h"
#include "sql-mailbox-list.h"

struct sql_list_iterate_context {
	struct mailbox_list_iterate_context ctx;
        struct imap_match_glob *glob;

	struct sql_result *result;
	struct mailbox_info info;
};

struct mailbox_list_iterate_context *
sql_mailbox_list_iter_init(struct mailbox_list *_list,
			   const char *const *patterns,
			   enum mailbox_list_iter_flags flags)
{
	struct sql_mailbox_list *list = (struct sql_mailbox_list *)_list;
	struct sql_list_iterate_context *ctx;
	string_t *query;
	const char *p, *name;

	ctx = i_new(struct sql_list_iterate_context, 1);
	ctx->ctx.list = _list;
	ctx->ctx.flags = flags;
	ctx->glob = imap_match_init_multiple(default_pool, patterns, TRUE,
					     _list->hierarchy_sep);

	query = t_str_new(256);
	if ((flags & MAILBOX_LIST_ITER_SELECT_SUBSCRIBED) != 0)
		str_append(query, "SELECT name FROM mailbox_subscriptions ");
	else
		str_append(query, "SELECT name FROM mailboxes ");

	str_printfa(query, "WHERE user_id = %u", list->user_id);

	for (; *patterns != NULL; patterns++) {
		for (p = *patterns; *p != '\0'; p++) {
			if (*p == '*' || *p == '%')
				break;
		}
		name = t_strdup_until(*patterns, p);
		str_printfa(query, " AND name LIKE '%s%%'",
			    sql_escape_string(list->db, name));
	}
	str_append(query, " ORDER BY name");

	ctx->result = sql_query_s(list->db, str_c(query));
	return &ctx->ctx;
}

int sql_mailbox_list_iter_deinit(struct mailbox_list_iterate_context *_ctx)
{
	struct sql_list_iterate_context *ctx =
		(struct sql_list_iterate_context *)_ctx;
	int ret = _ctx->failed ? -1 : 0;

	imap_match_deinit(&ctx->glob);
	sql_result_free(ctx->result);
	i_free(ctx);
	return ret;
}

const struct mailbox_info *
sql_mailbox_list_iter_next(struct mailbox_list_iterate_context *_ctx)
{
	struct sql_list_iterate_context *ctx =
		(struct sql_list_iterate_context *)_ctx;
	int ret;

	do {
		if ((ret = sql_result_next_row(ctx->result)) <= 0) {
			if (ret < 0) {
				_ctx->failed = TRUE;
				mailbox_list_set_critical(_ctx->list,
					"sql_mailbox_list query failed: %s",
					sql_result_get_error(ctx->result));
			}
			return NULL;
		}
		ctx->info.name = sql_result_get_field_value(ctx->result, 0);
	} while (imap_match(ctx->glob, ctx->info.name) != IMAP_MATCH_YES);

	ctx->info.flags = 0;
	return &ctx->info;
}
