/* Copyright (C) 2007 Timo Sirainen */

#include "lib.h"
#include "sql-api.h"
#include "sql-mailbox-list.h"

extern struct mailbox_list sql_mailbox_list;

static struct mailbox_list *sql_mailbox_list_alloc(void)
{
	struct sql_mailbox_list *list;
	pool_t pool;

	pool = pool_alloconly_create("fs mailbox list", 512);

	list = p_new(pool, struct sql_mailbox_list, 1);
	list->list = sql_mailbox_list;
	list->list.pool = pool;
	return &list->list;
}

static void sql_mailbox_list_deinit(struct mailbox_list *_list)
{
	struct sql_mailbox_list *list = (struct sql_mailbox_list *)_list;

	pool_unref(&list->list.pool);
}

static bool
sql_mailbox_list_is_valid_mask(struct mailbox_list *list ATTR_UNUSED,
			       const char *mask ATTR_UNUSED)
{
	return TRUE;
}

static bool
sql_mailbox_list_is_valid_existing_name(struct mailbox_list *list ATTR_UNUSED,
					const char *name ATTR_UNUSED)
{
	return TRUE;
}

static bool
sql_mailbox_list_is_valid_create_name(struct mailbox_list *list
				      	ATTR_UNUSED,
				      const char *name)
{
	if (*name == '\0' || strlen(name) > SQL_MAX_CREATE_MAILBOX_NAME_LENGTH)
		return FALSE;
	return TRUE;
}

static const char *
sql_mailbox_list_get_path(struct mailbox_list *list ATTR_UNUSED,
			  const char *name ATTR_UNUSED,
			  enum mailbox_list_path_type type ATTR_UNUSED)
{
	/* SQL has no file paths */
	return "";
}

static int
sql_mailbox_list_get_mailbox_name_status(struct mailbox_list *_list,
					 const char *name,
					 enum mailbox_name_status *status)
{
	struct sql_mailbox_list *list = (struct sql_mailbox_list *)_list;
	struct sql_result *result;
	const char *value;
	int ret = 0;

	result = sql_query_s(list->db, t_strdup_printf(
		"SELECT count(*) FROM mailboxes "
		"WHERE user_id = %u AND name = '%s'",
		list->user_id, sql_escape_string(list->db, name)));

	if (sql_result_next_row(result) < 0) {
		mailbox_list_set_critical(_list,
			"sql_get_mailbox_name_status query failed: %s",
			sql_result_get_error(result));
		value = NULL;
		ret = -1;
	} else {
		value = sql_result_get_field_value(result, 0);
	}
	sql_result_free(result);
	if (ret < 0)
		return -1;

	if (*value == '1') {
		*status = MAILBOX_NAME_EXISTS;
		return 0;
	}

	if (!mailbox_list_is_valid_create_name(_list, name)) {
		*status = MAILBOX_NAME_INVALID;
		return 0;
	}

	*status = MAILBOX_NAME_VALID;
	return 0;
}

static const char *
sql_mailbox_list_get_temp_prefix(struct mailbox_list *_list ATTR_UNUSED)
{
	return "";
}

static int sql_mailbox_list_set_subscribed(struct mailbox_list *_list,
					   const char *name, bool set)
{
	struct sql_mailbox_list *list = (struct sql_mailbox_list *)_list;

	return -1;
}

static int
sql_mailbox_list_delete_mailbox(struct mailbox_list *_list, const char *name)
{
	struct sql_mailbox_list *list = (struct sql_mailbox_list *)_list;
	struct sql_transaction_context *trans;
	const char *error;

	trans = sql_transaction_begin(list->db);
	sql_update(trans, t_strdup_printf(
		"DELETE FROM mailboxes WHERE user_id = %u AND name = '%s'",
		list->user_id, sql_escape_string(list->db, name)));
	if (sql_transaction_commit_s(&trans, &error) < 0) {
		mailbox_list_set_critical(_list,
			"sql_mailbox_delete query failed: %s", error);
		return -1;
	}
	return 0;
}

static int
sql_mailbox_list_rename_mailbox(struct mailbox_list *_list,
				const char *oldname, const char *newname)
{
	struct sql_mailbox_list *list = (struct sql_mailbox_list *)_list;
	struct sql_transaction_context *trans;
	const char *error;

	trans = sql_transaction_begin(list->db);
	sql_update(trans, t_strdup_printf(
		"UPDATE mailboxes SET name = '%s' "
		"WHERE user_id = %u AND name = '%s'",
		sql_escape_string(list->db, newname), list->user_id,
		sql_escape_string(list->db, oldname)));

	if (sql_transaction_commit_s(&trans, &error) < 0) {
		mailbox_list_set_critical(_list,
			"sql_rename_mailbox query failed: %s", error);
		return -1;
	}
	return 0;
}

struct mailbox_list sql_mailbox_list = {
	MEMBER(name) "sql",
	MEMBER(hierarchy_sep) '/',
	MEMBER(mailbox_name_max_length) PATH_MAX,

	{
		sql_mailbox_list_alloc,
		sql_mailbox_list_deinit,
		sql_mailbox_list_is_valid_mask,
		sql_mailbox_list_is_valid_existing_name,
		sql_mailbox_list_is_valid_create_name,
		sql_mailbox_list_get_path,
		sql_mailbox_list_get_mailbox_name_status,
		sql_mailbox_list_get_temp_prefix,
		NULL,
		sql_mailbox_list_iter_init,
		sql_mailbox_list_iter_next,
		sql_mailbox_list_iter_deinit,
		NULL,
		sql_mailbox_list_set_subscribed,
		sql_mailbox_list_delete_mailbox,
		sql_mailbox_list_rename_mailbox
	}
};
