/* Copyright (C) 2006 Timo Sirainen */

#include "lib.h"
#include "ioloop.h"
#include "istream.h"
#include "str.h"
#include "index-mail.h"
#include "sql-sync.h"
#include "sql-storage.h"
#include "sql-storage-query.h"

struct sql_save_context {
	struct mail_save_context ctx;

	struct sql_mailbox *mbox;
	struct mail_index_transaction *trans;
	struct sql_transaction_context *sql_trans;

	struct sql_sync_context *sync_ctx;

	/* updated for each appended mail: */
	uint32_t seq;
	struct istream *input;
	struct ostream *output;
	struct mail *mail;

	string_t *query;

	unsigned int in_header:1;
	unsigned int last_lf:1;
	unsigned int failed:1;
	unsigned int finished:1;
};

int sql_save_init(struct mailbox_transaction_context *_t,
		  enum mail_flags flags, struct mail_keywords *keywords,
		  time_t received_date, int timezone_offset ATTR_UNUSED,
		  const char *from_envelope ATTR_UNUSED,
		  struct istream *input, struct mail *dest_mail,
		  struct mail_save_context **ctx_r)
{
	struct sql_transaction_context *t =
		(struct sql_transaction_context *)_t;
	struct sql_mailbox *mbox = (struct sql_mailbox *)t->ictx.ibox;
	struct sql_save_context *ctx = t->save_ctx;
	const struct var_expand_table *variables;
	enum mail_flags save_flags;

	i_assert((t->ictx.flags & MAILBOX_TRANSACTION_FLAG_EXTERNAL) != 0);

	if (received_date == (time_t)-1)
		received_date = ioloop_time;

	if (ctx == NULL) {
		ctx = t->save_ctx = i_new(struct sql_save_context, 1);
		ctx->ctx.transaction = &t->ictx.mailbox_ctx;
		ctx->mbox = mbox;
		ctx->trans = t->ictx.trans;
		ctx->sql_trans = sql_transaction_begin(mbox->storage->db);

		ctx->query = str_new(default_pool, 65536);
	}
	ctx->input = input;

	/* add to index */
	/*save_flags = (flags & ~MAIL_RECENT) | MAIL_RECENT;
	mail_index_append(ctx->trans, 0, &ctx->seq);
	mail_index_update_flags(ctx->trans, ctx->seq, MODIFY_REPLACE,
				save_flags);
	if (keywords != NULL) {
		mail_index_update_keywords(ctx->trans, ctx->seq,
					   MODIFY_REPLACE, keywords);
	}*/

	/*if (dest_mail == NULL) {
		if (ctx->mail == NULL)
			ctx->mail = index_mail_alloc(_t, 0, NULL);
		dest_mail = ctx->mail;
	}
	if (mail_set_seq(dest_mail, ctx->seq) < 0)
		i_unreached();*/

	variables = var_expand_table_build('m', dec2str(ctx->mbox->id),
					   '\0');
	sql_storage_update(ctx->sql_trans, SQ_MAILBOX_SAVE_INIT, variables);

	str_truncate(ctx->query, 0);
	str_printfa(ctx->query, "INSERT INTO message_bodies "
		    "(message_id, header, body) VALUES "
		    "(currval('messages_id_seq'), '");

	ctx->in_header = TRUE;
	ctx->last_lf = TRUE;

	*ctx_r = &ctx->ctx;
	return ctx->failed ? -1 : 0;
}

static void append_escaped(struct sql_save_context *ctx,
			   const unsigned char *data, size_t size)
{
	const char *escaped;

	t_push();
	escaped = sql_escape_string(ctx->mbox->storage->db,
				    t_strndup(data, size));
	str_append(ctx->query, escaped);
	t_pop();
}

int sql_save_continue(struct mail_save_context *_ctx)
{
	struct sql_save_context *ctx = (struct sql_save_context *)_ctx;
	const unsigned char *data;
	size_t i, size;

	if (ctx->failed)
		return -1;

	/* FIXME: not a good idea to read in memory.. */
	data = i_stream_get_data(ctx->input, &size);
	if (size == 0)
		return 0;

	if (ctx->in_header) {
		size_t hdr_size = 0;

		for (i = 0; i < size; i++) {
			if (data[i] == '\r')
				;
			else if (data[i] == '\n') {
				if (ctx->last_lf) {
					hdr_size = i + 1;
					break;
				}
				ctx->last_lf = TRUE;
			} else {
				ctx->last_lf = FALSE;
			}
		}

		if (hdr_size > 0) {
			/* end of headers */
			append_escaped(ctx, data, hdr_size);
			str_append(ctx->query, "', '");
			i_stream_skip(ctx->input, hdr_size);
			data = i_stream_get_data(ctx->input, &size);
			ctx->in_header = FALSE;
		}
	}
	append_escaped(ctx, data, size);
	i_stream_skip(ctx->input, size);
	return 0;
}

int sql_save_finish(struct mail_save_context *_ctx)
{
	struct sql_save_context *ctx = (struct sql_save_context *)_ctx;

	ctx->finished = TRUE;

	if (ctx->in_header)
		str_append(ctx->query, "','");

	str_append(ctx->query, "')");
	sql_update(ctx->sql_trans, str_c(ctx->query));

	return ctx->failed ? -1 : 0;
}

void sql_save_cancel(struct mail_save_context *_ctx)
{
	struct sql_save_context *ctx = (struct sql_save_context *)_ctx;

	ctx->failed = TRUE;
	(void)sql_save_finish(_ctx);
}

int sql_transaction_save_commit_pre(struct sql_save_context *ctx)
{
	const struct mail_index_header *hdr;
	uint32_t uid, next_uid;
	const char *error;

	i_assert(ctx->finished);

	if (sql_transaction_commit_s(&ctx->sql_trans, &error) < 0) {
		mail_storage_set_critical(&ctx->mbox->storage->storage,
			"sql save commit failed: %s", error);
		return -1;
	}

	/*if (sql_mailbox_sync_begin(ctx->mbox, &ctx->sync_ctx) < 0) {
		ctx->failed = TRUE;
		sql_transaction_save_rollback(ctx);
		return -1;
	}

	hdr = mail_index_get_header(ctx->sync_ctx->sync_view);
	uid = hdr->next_uid;
	mail_index_append_assign_uids(ctx->trans, uid, &next_uid);*/

	return 0;
}

void sql_transaction_save_commit_post(struct sql_save_context *ctx)
{
	//(void)sql_mailbox_sync_finish(&ctx->sync_ctx, TRUE);
	sql_transaction_save_rollback(ctx);
}

void sql_transaction_save_rollback(struct sql_save_context *ctx)
{
	if (!ctx->finished)
		sql_save_cancel(&ctx->ctx);

	if (ctx->sql_trans != NULL)
		sql_transaction_rollback(&ctx->sql_trans);
	/*if (ctx->sync_ctx != NULL)
		(void)sql_mailbox_sync_finish(&ctx->sync_ctx, FALSE);*/

	if (ctx->mail != NULL)
		index_mail_free(ctx->mail);
	i_free(ctx);
}
