/* Copyright (C) 2007 Timo Sirainen */

#include "lib.h"
#include "str.h"
#include "sql-storage.h"
#include "sql-storage-query.h"

struct sql_storage_query sql_storage_queries[SQ_COUNT] = {
	{
		"mailbox_sync",
		"SELECT next_uid, next_modseq, message_count "
		"FROM mailboxes WHERE id = %m"
	},
	{
		"mailbox_sync_newmsgs",
		"SELECT * "
		"FROM messages WHERE mailbox_id = %m "
		"AND uid >= %u AND uid <= %w"
	},
	{
		"mailbox_sync_list_uids",
		"SELECT uid FROM messages WHERE mailbox_id = %m "
		"AND uid <= %w "
		"ORDER BY uid"
	},
	{
		"mailbox_sync_list_flags",
		"SELECT uid, seen, answered, flagged, deleted, draft "
		"FROM messages WHERE mailbox_id = %m "
		"AND uid <= %w AND modseq >= %s "
		"ORDER BY uid"
	},
	{
		"mailbox_sync_expunge",
		"DELETE FROM messages WHERE mailbox_id = %m "
		"AND uid >= %u AND uid <= %w"
	},
	{
		"mailbox_sync_update",
		"UPDATE messages SET %c WHERE mailbox_id = %m "
		"AND uid >= %u AND uid <= %w"
	},
	{
		"mailbox_save_init",
		"INSERT INTO messages (mailbox_id) VALUES (%m)"
	},
	{
		"mailbox_body_init",
		"SELECT m.id, "
		"  octet_length(mb.header) + octet_length(mb.body) AS len "
		"FROM messages AS m "
		"JOIN message_bodies AS mb ON (m.id = mb.message_id) "
		"WHERE m.mailbox_id = %m AND m.uid = %u"
	},
	{
		"mailbox_body_template",
		"SELECT substring(header || body from (%%o+1) for %%s) "
		"FROM message_bodies WHERE message_id = %i"
	}

};

struct sql_result *
sql_storage_query(struct sql_storage *storage, enum sql_storage_query_id id,
		  const struct var_expand_table *variables)
{
	string_t *query;
	struct sql_result *result;

	t_push();
	query = t_str_new(256);
	var_expand(query, sql_storage_queries[id].query, variables);

	result = sql_query_s(storage->db, str_c(query));
	t_pop();
	return result;
}

int sql_storage_query_row(struct sql_storage *storage,
			  enum sql_storage_query_id id,
			  const struct var_expand_table *variables,
			  const struct sql_field_def *fields,
			  void *dest, size_t dest_size,
			  struct sql_result **result_r)
{
	struct sql_result *result;
	int ret;

	result = sql_storage_query(storage, id, variables);
	sql_result_setup_fetch(result, fields, dest, dest_size);
	if ((ret = sql_result_next_row(result)) < 0)
		sql_storage_query_failed(storage, id, result);

	if (ret <= 0)
		result = NULL;
	else if (result_r == NULL) {
		for (; fields->name != NULL; fields++)
			i_assert(fields->type != SQL_TYPE_STR);
		sql_result_free(result);
	}
	if (result_r != NULL)
		*result_r = result;
	return ret;
}

void sql_storage_update(struct sql_transaction_context *trans,
			enum sql_storage_query_id id,
			const struct var_expand_table *variables)
{
	string_t *query;

	t_push();
	query = t_str_new(256);
	var_expand(query, sql_storage_queries[id].query, variables);
	sql_update(trans, str_c(query));
	t_pop();
}

void sql_storage_query_failed(struct sql_storage *storage,
			      enum sql_storage_query_id id,
			      struct sql_result *result)
{
	mail_storage_set_critical(&storage->storage, "SQL query %s failed: %s",
				  sql_storage_queries[id].name,
				  sql_result_get_error(result));
}
