/* Copyright (C) 2006 Timo Sirainen */

#include "lib.h"
#include "sql-api.h"
#include "index-mail.h"
#include "mail-copy.h"
#include "sql-sync.h"
#include "sql-mailbox-list.h"
#include "sql-storage.h"

#include <stdlib.h>

struct mailbox_open_query_row {
	unsigned long long id;
	unsigned int uid_validity;
};

static struct sql_field_def mailbox_open_query_row_defs[] = {
	SQL_DEF_STRUCT_ULLONG(id, mailbox_open_query_row),
	SQL_DEF_STRUCT_UINT(uid_validity, mailbox_open_query_row),

	{ 0, NULL, 0 }
};

extern struct mail_storage sql_storage;
extern struct mailbox sql_mailbox;

static struct mail_storage *sql_storage_alloc(void)
{
	struct sql_storage *storage;
	pool_t pool;

	pool = pool_alloconly_create("sql storage", 512);
	storage = p_new(pool, struct sql_storage, 1);
	storage->storage = sql_storage;
	storage->storage.pool = pool;

	return &storage->storage;
}

static int sql_storage_create(struct mail_storage *_storage, const char *data,
			      const char **error_r)
{
	struct sql_storage *storage = (struct sql_storage *)_storage;
	struct sql_db *db;
	struct sql_result *result;
	struct mailbox_list_settings list_set;
	struct sql_mailbox_list *list;
	const char *p, *value, *error;
	int ret;

	*error_r = NULL;

	memset(&list_set, 0, sizeof(list_set));
	list_set.mail_storage_flags = &_storage->flags;
	list_set.lock_method = &_storage->lock_method;

	if (data == NULL || *data == '\0') {
		/* we won't do any guessing for this format. */
		*error_r = "mailbox location not given";
		return -1;
	}

	if ((_storage->flags & MAIL_STORAGE_FLAG_DEBUG) != 0)
		i_info("sql: data=%s", data);

	p = strchr(data, ':');
	if (p == NULL) {
		*error_r = "Missing ':' in connect string";
		return -1;
	}

	db = sql_init(t_strcut(data, ':'), p+1);
	if (db == NULL)
		i_error("Couldn't connect to database");
	result = sql_query_s(db, t_strdup_printf(
		"SELECT id FROM users WHERE name = '%s'",
		sql_escape_string(db, _storage->user)));
	if ((ret = sql_result_next_row(result)) <= 0) {
		if (ret < 0)
			error = sql_result_get_error(result);
		else
			error = "User not found";

		*error_r = t_strdup_printf("user lookup query failed: %s",
					   error);
		sql_result_free(result);
		sql_deinit(&db);
		return -1;
	}

	value = sql_result_get_field_value(result, 0);
	sql_result_free(result);

	if (mailbox_list_alloc("sql", &_storage->list, error_r) < 0) {
		sql_deinit(&db);
		return -1;
	}

	storage->db = db;
	storage->user_id = strtoul(value, NULL, 10);

	list = (struct sql_mailbox_list *)_storage->list;
	list->db = db;
	list->user_id = storage->user_id;

	mailbox_list_init(_storage->list, _storage->ns, &list_set,
			  mail_storage_get_list_flags(_storage->flags));
	return 0;
}

static void sql_storage_destroy(struct mail_storage *_storage)
{
	struct sql_storage *storage = (struct sql_storage *) _storage;

	sql_deinit(&storage->db);
}

static struct mailbox *
sql_mailbox_open(struct mail_storage *_storage, const char *name,
		 struct istream *input, enum mailbox_open_flags flags)
{
	struct sql_storage *storage = (struct sql_storage *)_storage;
	struct sql_result *result;
	struct mailbox_open_query_row row;
	struct sql_mailbox *mbox;
	struct mail_index *index;
	pool_t pool;
	int ret;

	if (input != NULL) {
		mail_storage_set_critical(_storage,
			"SQL storage doesn't support streamed mailboxes");
		return NULL;
	}

	result = sql_query_s(storage->db, t_strdup_printf(
		"SELECT id, uid_validity FROM mailboxes WHERE name = '%s'",
		sql_escape_string(storage->db, name)));
	sql_result_setup_fetch(result, mailbox_open_query_row_defs,
			       &row, sizeof(row));
	if ((ret = sql_result_next_row(result)) <= 0) {
		if (ret == 0) {
			mail_storage_set_error(_storage, MAIL_ERROR_NOTFOUND,
				t_strdup_printf(MAIL_ERRSTR_MAILBOX_NOT_FOUND,
						name));
		} else {
			mail_storage_set_critical(_storage,
				"sql_mailbox_create query failed: %s",
				sql_result_get_error(result));
		}
		sql_result_free(result);
		return NULL;
	}

	index = index_storage_alloc(_storage, name, flags, NULL);

	pool = pool_alloconly_create("sql mailbox", 1024);
	mbox = p_new(pool, struct sql_mailbox, 1);
	mbox->ibox.box = sql_mailbox;
	mbox->ibox.box.pool = pool;
	mbox->ibox.storage = &storage->storage;
	mbox->ibox.mail_vfuncs = &sql_mail_vfuncs;
	mbox->ibox.index = index;
	mbox->storage = storage;

	mbox->id = row.id;
	mbox->uid_validity = row.uid_validity;
	sql_result_free(result);

	index_storage_mailbox_init(&mbox->ibox, name, flags, FALSE);
	return &mbox->ibox.box;
}

static int sql_mailbox_create(struct mail_storage *_storage,
			      const char *name, bool directory ATTR_UNUSED)
{
	struct sql_storage *storage = (struct sql_storage *)_storage;
	struct sql_transaction_context *trans;
	const char *error;

	trans = sql_transaction_begin(storage->db);
	sql_update(trans, t_strdup_printf(
		"INSERT INTO mailboxes (user_id, name) VALUES (%u, '%s')",
		storage->user_id, sql_escape_string(storage->db, name)));

	if (sql_transaction_commit_s(&trans, &error) < 0) {
		mail_storage_set_critical(_storage,
			"sql_mailbox_create query failed: %s", error);
		return -1;
	}
	return 0;
}

static void sql_mailbox_notify_changes(struct mailbox *box ATTR_UNUSED)
{
}

static void sql_storage_class_init(void)
{
	mailbox_list_register(&sql_mailbox_list);
	sql_transaction_class_init();
}

static void sql_storage_class_deinit(void)
{
	sql_transaction_class_deinit();
	mailbox_list_unregister(&sql_mailbox_list);
}

struct mail_storage sql_storage = {
	MEMBER(name) SQL_STORAGE_NAME,
	MEMBER(mailbox_is_file) FALSE,

	{
		sql_storage_class_init,
		sql_storage_class_deinit,
		sql_storage_alloc,
		sql_storage_create,
		sql_storage_destroy,
		NULL,
		sql_mailbox_open,
		sql_mailbox_create
	}
};

struct mailbox sql_mailbox = {
	MEMBER(name) NULL, 
	MEMBER(storage) NULL, 

	{
		index_storage_is_readonly,
		index_storage_allow_new_keywords,
		index_storage_mailbox_close,
		index_storage_get_status,
		NULL,
		NULL,
		sql_mailbox_sync_init,
		index_mailbox_sync_next,
		index_mailbox_sync_deinit,
		NULL,
		sql_mailbox_notify_changes,
		index_transaction_begin,
		index_transaction_commit,
		index_transaction_rollback,
		index_keywords_create,
		index_keywords_free,
		index_storage_get_uids,
		index_mail_alloc,
		index_header_lookup_init,
		index_header_lookup_deinit,
		index_storage_search_init,
		index_storage_search_deinit,
		index_storage_search_next_nonblock,
		index_storage_search_next_update_seq,
		sql_save_init,
		sql_save_continue,
		sql_save_finish,
		sql_save_cancel,
		mail_storage_copy,
		index_storage_is_inconsistent
	}
};
