/* Copyright (C) 2007 Timo Sirainen */

#include "lib.h"
#include "ioloop.h"
#include "str.h"
#include "sql-storage.h"
#include "sql-storage-query.h"
#include "sql-sync.h"

#include <stdlib.h>

struct mailbox_sync_row {
	unsigned int next_uid;
	unsigned int message_count;
	unsigned long long next_modseq;
};

static struct sql_field_def mailbox_sync_row_defs[] = {
	SQL_DEF_STRUCT_UINT(next_uid, mailbox_sync_row),
	SQL_DEF_STRUCT_UINT(message_count, mailbox_sync_row),
	SQL_DEF_STRUCT_ULLONG(next_modseq, mailbox_sync_row),

	{ 0, NULL, 0 }
};

struct mailbox_sync_uid_row {
	unsigned int uid;
};

static struct sql_field_def mailbox_sync_uid_row_defs[] = {
	SQL_DEF_STRUCT_UINT(uid, mailbox_sync_uid_row),

	{ 0, NULL, 0 }
};

struct mailbox_sync_newmsgs_row {
	unsigned int uid;

	bool seen;
	bool answered;
	bool flagged;
	bool deleted;
	bool draft;
};

static struct sql_field_def mailbox_sync_newmsgs_row_defs[] = {
	SQL_DEF_STRUCT_UINT(uid, mailbox_sync_newmsgs_row),
	SQL_DEF_STRUCT_BOOL(seen, mailbox_sync_newmsgs_row),
	SQL_DEF_STRUCT_BOOL(answered, mailbox_sync_newmsgs_row),
	SQL_DEF_STRUCT_BOOL(flagged, mailbox_sync_newmsgs_row),
	SQL_DEF_STRUCT_BOOL(deleted, mailbox_sync_newmsgs_row),
	SQL_DEF_STRUCT_BOOL(draft, mailbox_sync_newmsgs_row),

	{ 0, NULL, 0 }
};

struct mail_flag_map {
	enum mail_flags flag;
	const char *name;
	size_t offset;
};
#define MAIL_FLAG_OFFSET(name) offsetof(struct mailbox_sync_newmsgs_row, name)
static const struct mail_flag_map mail_flag_map[] = {
	{ MAIL_SEEN, "seen", MAIL_FLAG_OFFSET(seen) },
	{ MAIL_ANSWERED, "answered", MAIL_FLAG_OFFSET(answered) },
	{ MAIL_FLAGGED, "flagged", MAIL_FLAG_OFFSET(flagged) },
	{ MAIL_DELETED, "deleted", MAIL_FLAG_OFFSET(deleted) },
	{ MAIL_DRAFT, "draft", MAIL_FLAG_OFFSET(draft) },
	{ 0, NULL, 0 }
};

static enum mail_flags
newmsgs_row_get_flags(const struct mailbox_sync_newmsgs_row *row)
{
	enum mail_flags flags = 0;
	const struct mail_flag_map *map;

	for (map = mail_flag_map; map->name != NULL; map++) {
		bool *set = STRUCT_MEMBER_P(row, map->offset);

		if (*set)
			flags |= map->flag;
	}
	return flags;
}

static int
sql_sync_from_database_new_msgs(struct sql_sync_context *ctx, uint32_t max_uid,
				unsigned int *new_msgs_count_r)
{
	const struct var_expand_table *variables;
	const struct mail_index_header *hdr;
	struct sql_result *result;
	struct mailbox_sync_newmsgs_row row;
	unsigned int count = 0;
	uint32_t seq;
	int ret;

	hdr = mail_index_get_header(ctx->sync_view);
	variables = var_expand_table_build('m', dec2str(ctx->mbox->id),
					   'u', dec2str(hdr->next_uid),
					   'w', dec2str(max_uid),
					   '\0');

	result = sql_storage_query(ctx->mbox->storage,
				   SQ_MAILBOX_SYNC_NEWMSGS, variables);
	sql_result_setup_fetch(result, mailbox_sync_newmsgs_row_defs,
			       &row, sizeof(row));
	while ((ret = sql_result_next_row(result)) > 0) {
		mail_index_append(ctx->trans, row.uid, &seq);
		mail_index_update_flags(ctx->trans, seq, MODIFY_REPLACE,
					newmsgs_row_get_flags(&row));
		count++;
	}
	if (ret < 0) {
		sql_storage_query_failed(ctx->mbox->storage,
					 SQ_MAILBOX_SYNC_NEWMSGS, result);
	}
	sql_result_free(result);

	*new_msgs_count_r = count;
	return ret;
}

static int sql_sync_from_database_expunges(struct sql_sync_context *ctx)
{
	const struct var_expand_table *variables;
	const struct mail_index_header *hdr;
	struct sql_result *result;
	struct mailbox_sync_uid_row row;
	uint32_t seq, uid;
	int ret = 1;

	hdr = mail_index_get_header(ctx->sync_view);
	variables = var_expand_table_build('m', dec2str(ctx->mbox->id),
					   'w', dec2str(hdr->next_uid - 1),
					   '\0');

	/* get a list of existing UIDs to find out what's missing */
	result = sql_storage_query(ctx->mbox->storage,
				   SQ_MAILBOX_SYNC_LIST_UIDS, variables);
	sql_result_setup_fetch(result, mailbox_sync_uid_row_defs,
			       &row, sizeof(row));
	seq = 1;
	while ((ret = sql_result_next_row(result)) > 0) {
		uid = 0;
		while (seq <= hdr->messages_count) {
			mail_index_lookup_uid(ctx->sync_view, seq, &uid);
			if (uid >= row.uid)
				break;
			mail_index_expunge(ctx->trans, seq);
		}
		if (uid > row.uid) {
			mail_storage_set_critical(&ctx->mbox->storage->storage,
				"UID inserted into middle of the mailbox");
			ret = -2;
			break;
		}
	}
	if (ret == -1) {
		sql_storage_query_failed(ctx->mbox->storage,
					 SQ_MAILBOX_SYNC_LIST_UIDS, result);
	}
	sql_result_free(result);
	return ret < 0 ? -1 : ret;
}

static int sql_sync_from_database_flag_changes(struct sql_sync_context *ctx)
{
	const struct var_expand_table *variables;
	const struct mail_index_header *hdr;
	struct sql_result *result;
	struct mailbox_sync_newmsgs_row row;
	uint32_t seq;
	int ret = 1;

	hdr = mail_index_get_header(ctx->sync_view);
	variables = var_expand_table_build('m', dec2str(ctx->mbox->id),
					   's', dec2str(hdr->sync_size),
					   'w', dec2str(hdr->next_uid - 1),
					   '\0');

	result = sql_storage_query(ctx->mbox->storage,
				   SQ_MAILBOX_SYNC_LIST_FLAGS, variables);
	sql_result_setup_fetch(result, mailbox_sync_newmsgs_row_defs,
			       &row, sizeof(row));
	seq = 1;
	while ((ret = sql_result_next_row(result)) > 0) {
		if (mail_index_lookup_seq(ctx->sync_view, row.uid, &seq)) {
			mail_index_update_flags(ctx->trans, seq, MODIFY_REPLACE,
						newmsgs_row_get_flags(&row));
		}
	}
	if (ret == -1) {
		sql_storage_query_failed(ctx->mbox->storage,
					 SQ_MAILBOX_SYNC_LIST_FLAGS, result);
	}
	sql_result_free(result);
	return ret < 0 ? -1 : ret;
}

static int sql_sync_from_database(struct sql_sync_context *ctx)
{
	struct mail_storage *storage = &ctx->mbox->storage->storage;
	const struct var_expand_table *variables;
	const struct mail_index_header *hdr;
	struct mailbox_sync_row row;
	unsigned int new_msgs_count = 0;
	int ret;

	hdr = mail_index_get_header(ctx->sync_view);
	if (hdr->uid_validity == 0) {
		uint32_t uid_validity = ctx->mbox->uid_validity;

		mail_index_update_header(ctx->trans,
			offsetof(struct mail_index_header, uid_validity),
			&uid_validity, sizeof(uid_validity), TRUE);
	}

	variables = var_expand_table_build('m', dec2str(ctx->mbox->id),
					   '\0');
	ret = sql_storage_query_row(ctx->mbox->storage, SQ_MAILBOX_SYNC,
				    variables, mailbox_sync_row_defs,
				    &row, sizeof(row), NULL);
	if (ret <= 0) {
		if (ret == 0)
			mailbox_set_deleted(&ctx->mbox->ibox.box);
		return -1;
	}

	if (hdr->next_uid != row.next_uid) {
		/* new messages have been added since last check */
		if (hdr->next_uid > row.next_uid) {
			mail_storage_set_critical(storage,
				"next_uid lowered unexpectedly");
			return -1;
		}

		if (sql_sync_from_database_new_msgs(ctx, row.next_uid - 1,
						    &new_msgs_count) < 0)
			return -1;

		mail_index_update_header(ctx->trans,
			offsetof(struct mail_index_header, next_uid),
			&row.next_uid, sizeof(row.next_uid), FALSE);
	}

	if (hdr->messages_count + new_msgs_count != row.message_count) {
		/* some messages have been expunged */
		if (sql_sync_from_database_expunges(ctx) < 0)
			return -1;
	}

	if (hdr->sync_size != row.next_modseq) {
		/* flags have changed */
		if (sql_sync_from_database_flag_changes(ctx) < 0)
			return -1;

		mail_index_update_header(ctx->trans,
			offsetof(struct mail_index_header, sync_size),
			&row.next_modseq, sizeof(row.next_modseq), FALSE);
	}

	return 0;
}

static const char *
build_flag_changes_update(const struct mail_index_sync_rec *sync_rec)
{
	const struct mail_flag_map *map;
	string_t *str;
	char value = '\0';

	str = t_str_new(128);
	for (map = mail_flag_map; map->name != NULL; map++) {
		if ((sync_rec->add_flags & map->flag) != 0)
			value = 't';
		else if ((sync_rec->remove_flags & map->flag) != 0)
			value = 'f';
		else
			continue;

		if (str_len(str) > 0)
			str_append_c(str, ',');
		str_printfa(str, "%s='%c'", map->name, value);
	}
	return str_c(str);
}

static int sql_sync_record(struct sql_sync_context *ctx,
			   const struct mail_index_sync_rec *sync_rec)
{
	const struct var_expand_table *variables;
	const char *changes;

	t_push();
	switch (sync_rec->type) {
	case MAIL_INDEX_SYNC_TYPE_EXPUNGE:
		variables = var_expand_table_build('m', dec2str(ctx->mbox->id),
						   'u', dec2str(sync_rec->uid1),
						   'w', dec2str(sync_rec->uid2),
						   '\0');
		sql_storage_update(ctx->sql_trans, SQ_MAILBOX_SYNC_EXPUNGE,
				   variables);
		break;
	case MAIL_INDEX_SYNC_TYPE_FLAGS:
		changes = build_flag_changes_update(sync_rec);
		if (*changes == '\0')
			break;

		variables = var_expand_table_build('m', dec2str(ctx->mbox->id),
						   'u', dec2str(sync_rec->uid1),
						   'w', dec2str(sync_rec->uid2),
						   'c', changes,
						   '\0');
		sql_storage_update(ctx->sql_trans, SQ_MAILBOX_SYNC_UPDATE,
				   variables);
		break;
	default:
		break;
	}
	t_pop();
	return 0;
}

static int sql_sync_to_database(struct sql_sync_context *ctx)
{
	struct mail_index_sync_rec sync_rec;
	int ret;

	while ((ret = mail_index_sync_next(ctx->index_sync_ctx,
					   &sync_rec)) > 0) {
		if (sql_sync_record(ctx, &sync_rec) < 0)
			return -1;
	}
	return 0;
}

int sql_mailbox_sync_begin(struct sql_mailbox *mbox,
			   struct sql_sync_context **ctx_r)
{
	struct sql_sync_context *ctx;
	enum mail_index_sync_flags sync_flags = 0;

	ctx = i_new(struct sql_sync_context, 1);
	ctx->mbox = mbox;

	if (!mbox->ibox.keep_recent)
		sync_flags |= MAIL_INDEX_SYNC_FLAG_DROP_RECENT;

	if (mail_index_sync_begin(mbox->ibox.index, &ctx->index_sync_ctx,
				  &ctx->sync_view, &ctx->trans,
				  sync_flags) < 0) {
		mail_storage_set_index_error(&mbox->ibox);
		i_free(ctx);
		return -1;
	}
	ctx->sql_trans = sql_transaction_begin(mbox->storage->db);

	if (sql_sync_from_database(ctx) < 0 ||
	    sql_sync_to_database(ctx) < 0) {
		mail_index_sync_rollback(&ctx->index_sync_ctx);
		i_free(ctx);
		return -1;
	}

	*ctx_r = ctx;
	return 0;
}

int sql_mailbox_sync_finish(struct sql_sync_context **_ctx, bool success)
{
	struct sql_sync_context *ctx = *_ctx;
	const char *error;
	int ret = success ? 0 : -1;

	if (ret == 0) {
		if (sql_transaction_commit_s(&ctx->sql_trans, &error) < 0) {
			mail_storage_set_critical(&ctx->mbox->storage->storage,
				"sql sync commit failed: %s", error);
			ret = -1;
		}
	} else {
		sql_transaction_rollback(&ctx->sql_trans);
	}

	*_ctx = NULL;
	if (ret == 0) {
		if (mail_index_sync_commit(&ctx->index_sync_ctx) < 0) {
			mail_storage_set_index_error(&ctx->mbox->ibox);
			ret = -1;
		}
	} else {
		mail_index_sync_rollback(&ctx->index_sync_ctx);
	}
	i_free(ctx);
	return 0;
}

int sql_mailbox_sync(struct sql_mailbox *mbox)
{
	struct sql_sync_context *sync_ctx;

	if (sql_mailbox_sync_begin(mbox, &sync_ctx) < 0)
		return -1;

	return sql_mailbox_sync_finish(&sync_ctx, TRUE);
}

struct mailbox_sync_context *
sql_mailbox_sync_init(struct mailbox *box, enum mailbox_sync_flags flags)
{
	struct sql_mailbox *mbox = (struct sql_mailbox *)box;
	int ret = 0;

	if (!box->opened)
		index_storage_mailbox_open(&mbox->ibox);

	if ((flags & MAILBOX_SYNC_FLAG_FAST) == 0 ||
	    mbox->ibox.sync_last_check + MAILBOX_FULL_SYNC_INTERVAL <=
	    ioloop_time)
		ret = sql_mailbox_sync(mbox);

	return index_mailbox_sync_init(box, flags, ret < 0);
}
